/*******************************************************************************
 * LinkedList
 *
 * A simple doubly linked list class in C.
 *
 * Not thread safe.
 *
 ******************************************************************************/

#include "LinkedList.h"

#include <stdlib.h>

/*******************************************************************************
 * "Private" inner class.
 */ 
struct ll_node
   {
   ll_node *	next;
   ll_node *	prev;
   void *	data;
   };


/*******************************************************************************
 * Insert an element at end of list.
 * Sets curr and tail pointers to the new node.
 * Returns data on success, NULL on error.
 */
void * ll_push(LinkedList * this, void * data)
   {
   ll_node * new_node;

   if ((!this) || (!data))
      return NULL;

   if (!(new_node = calloc(1, sizeof(ll_node))))
      return NULL;

   new_node->data = data;

   if (!(this->Tail))
      {
      /* empty list */
      this->Head = this->Tail = this->Curr = new_node;
      }
   else
      {
      new_node->prev = this->Tail;
      this->Tail->next = new_node;
      this->Tail = this->Curr = new_node;
      }

   this->Size++;
   return this->Curr->data;
   }


/*******************************************************************************
 * Insert an element after the one pointed to by this->curr.
 * Sets curr pointer to the new node (and tail if curr = tail when called).
 * Returns data on success, NULL on error.
 */
void * ll_insert(LinkedList * this, void * data)
   {
   ll_node * new_node;

   if ((!this) || (!data))
      return NULL;

   if (this->Curr == this->Tail)
      {
      return ll_push(this, data);
      }

   if (!(new_node = calloc(1, sizeof(ll_node))))
      return NULL;

   new_node->data = data;
   new_node->prev = this->Curr;
   new_node->next = this->Curr->next;
   this->Curr->next = new_node;
   new_node->next->prev = new_node;
   this->Curr = new_node;

   this->Size++;
   return this->Curr->data;
   }


/*******************************************************************************
 * Insert an element and make it the new head.
 * Sets head and curr pointers to the new node.
 * Returns data on success, NULL on error.
 */
void * ll_insert_at_head(LinkedList * this, void * data)
   {
   ll_node * new_node;

   if ((!this) || (!data))
      return NULL;

   if (!(new_node = calloc(1, sizeof(ll_node))))
      return NULL;

   new_node->data = data;
   new_node->prev = NULL;
   new_node->next = this->Head;;

   if (this->Head)
      this->Head->prev = new_node;
   this->Head = this->Curr = new_node;
   
   this->Size++;
   return this->Curr->data;
   }


/*******************************************************************************
 * Remove the tail element from the list.
 * Sets curr to tail.
 * Return data, or NULL if the list is empty.
 */
void * ll_pop(LinkedList * this)
   {
   void * data;
   ll_node * tmp;

   if (!this)
      return NULL;

   if (!(this->Tail))
      return NULL;

   tmp = this->Tail;
   data = this->Tail->data;

   if (!(this->Tail->prev))
      {
      /* Only element in list */
      this->Head = this->Tail = this->Curr = NULL;
      }
   else
      {
      this->Tail->prev->next = NULL;
      this->Tail = this->Curr = this->Tail->prev;
      }
   
   free(tmp);
   this->Size--;

   return data;
   }


/*******************************************************************************
 * Delete the element pointed to by this->curr.
 * A pointer to the stored data is returned, or NULL is no list or an empty list
 * is specified.  It must be freed by the caller.
 */
void * ll_remove(LinkedList * this)
   {
   void * data;
   ll_node * tmp;

   if (!this)
      return NULL;

   if (this->Curr == this->Tail)
      return ll_pop(this);

   tmp = this->Curr;
   data = this->Curr->data;

   if (this->Curr == this->Head)
      {
      this->Curr->next->prev = NULL;
      this->Head = this->Curr = this->Head->next;
      }
   else
      {
      this->Curr->prev->next = this->Curr->next;
      this->Curr->next->prev = this->Curr->prev;
      }

   free(tmp);
   this->Size--;

   return data;
   }


/*******************************************************************************
 * Return the size of the list, or -1 is list is null.
 */
int ll_size(LinkedList * this)
   {
   if (!this) return -1;

   return this->Size;
   }


/*******************************************************************************
 * Select the first element in the list if there is one.
 * Return data pointer of new current element, or NULL is list is empty.
 */
void * ll_head(LinkedList * this)
   {
   if (!this) return NULL;

   this->Curr = this->Head;

   if (this->Curr)
      return this->Curr->data;
   else
      return NULL;
   }


/*******************************************************************************
 * Select the last element in the list if there is one.
 * Return data pointer of new current element, or NULL is list is empty.
 */
void * ll_tail(LinkedList * this)
   {
   if (!this) return NULL;
   
   this->Curr = this->Tail;

   if (this->Curr)
      return this->Curr->data;
   else
      return NULL;
   }


/*******************************************************************************
 * Select the next element in the list if there is one.
 * Return data pointer of new current element, or NULL is list is empty.
 */
void * ll_next(LinkedList * this)
   {
   if (!this) return NULL;
   
   if (this->Curr->next)
      this->Curr = this->Curr->next;

   if (this->Curr)
      return this->Curr->data;
   else
      return NULL;
   }


/*******************************************************************************
 * Select previous element in the list if there is one.
 * Return data pointer of new current element, or NULL is list is empty.
 */
void * ll_prev(LinkedList * this)
   {
   if (!this) return NULL;

   if (this->Curr->prev)
      this->Curr = this->Curr->prev;

   if (this->Curr)
      return this->Curr->data;
   else
      return NULL;
   }


/*******************************************************************************
 * Return the data pointer of the current node, or NULL is list is empty.
 */
void * ll_data(LinkedList * this)
   {
   if (!this) return NULL;

   if (this->Curr)
      return this->Curr->data;
   else
      return NULL;
   }


/*******************************************************************************
 * Constuctor.   Call this when creating a new list.
 */ 
void ll_init(LinkedList * this)
   {
   if (this)
      {
      this->push   = ll_push;
      this->pop    = ll_pop;
      this->insert = ll_insert;
      this->remove = ll_remove;
      this->insert_at_head = ll_insert_at_head;
      this->size   = ll_size;
      this->head   = ll_head;
      this->tail   = ll_tail;
      this->next   = ll_next;
      this->prev   = ll_prev;
      this->data   = ll_data;

      this->Head = this->Tail = this->Curr = NULL;
      this->Size = 0;
      }
   }


